import ConfigParser
import os
import sys
import os
import logging
from logging.handlers import SysLogHandler as sl

# FIXME document everything

def _get_prefix(): return os.path.abspath(os.path.join(os.path.dirname(sys.argv[0]),".."))

def _get_data_dir():
	prefix = _get_prefix()
	if prefix == "/usr": return "/var"
	return os.path.join(prefix,"var")

def _get_sysconfdir():
	prefix = _get_prefix()
	if prefix == "/usr": return "/etc"
	return os.path.join(prefix,"etc")

def _get_userconfdir():
	home = None
	try: home = os.environ["HOME"]
	except: pass
	return home

_config_filename = "metadata-service.conf"

def _get_configfile():
	userconfdir = _get_userconfdir()
	if userconfdir:
		sysconffile = os.path.join(userconfdir,"."+_config_filename)
		if os.path.exists(sysconffile): return sysconffile
	paths = [_get_sysconfdir(),"/usr/local/etc","/etc"]
	for path in paths:
		sysconffile = os.path.join(path,_config_filename)
		if os.path.exists(sysconffile): return sysconffile

def detect_config_file():
	return _get_configfile()

class MetadataServiceConfig(ConfigParser.ConfigParser):
 
	_config_filename = None
	_data_dir = None

 	def __init__(self,filename=None):
 		ConfigParser.ConfigParser.__init__(self)
		if filename: self._config_filename = os.path.abspath(filename)
		else: self._config_filename = detect_config_file()
		if self._config_filename: self.read(self._config_filename)
 		self.__apply_defaults()
 
	def get_filename(self):
		return self._config_filename

	def get_data_dir(self):
		return self.get("service","root directory")

	def get_socket(self):
		# FIXME: make the data dir not relative to the binary (which may be located in different places) but relative to this file itself.
		return self.get("socket interface","socket")

	def get_lock_file(self):
		# FIXME: make the data dir not relative to the binary (which may be located in different places) but relative to this file itself.
		return self.get("service","lock file")

 	def __apply_defaults(self):
		if not self.has_section("service"): self.add_section("service")

 		try: data_dir = self.get("service","root directory")
 		except ConfigParser.NoOptionError:
			self.set("service","root directory",os.path.join(_get_data_dir(),"lib","metadata-service"))

 		try: dbdir = self.get("service","lock file")
 		except ConfigParser.NoOptionError:
 			dbdir = os.path.join(self.get_data_dir(),"lock")
			self.set("service","lock file",dbdir)

		if not self.has_section("index"): self.add_section("index")
 		try: dbdir = self.get("index","path")
 		except ConfigParser.NoOptionError:
 			dbdir = os.path.join(self.get_data_dir(),"index")
			self.set("index","path",dbdir)

		if not self.has_section("logging"): self.add_section("logging")
 		try:
			level = self.get("logging","log level")
			l = { "INFO":logging.INFO,
				"DEBUG":logging.DEBUG,
				"ERROR":logging.ERROR,
				"CRITICAL":logging.CRITICAL,
				"WARNING":logging.WARNING
			    }
			self.set("logging","log level",str(l[level]))
 		except ConfigParser.NoOptionError: self.set("logging","log level",str(logging.INFO))
		except KeyError: raise ValueError, "Invalid log level %s"%level

		try: 
			syslog = self.getboolean("logging","use syslog")
		except ConfigParser.NoOptionError: 
			self.set("logging","use syslog","True")

 		try:
			facility = self.get("logging","log facility")
			facilities = {'ALERT':sl.LOG_ALERT,
				'AUTH':sl.LOG_AUTH, 'AUTHPRIV':sl.LOG_AUTHPRIV,
				'CRON':sl.LOG_CRON, 'DAEMON':sl.LOG_DAEMON,
				'KERN':sl.LOG_KERN, 'LOCAL0':sl.LOG_LOCAL0,
				'LOCAL1':sl.LOG_LOCAL1, 'LOCAL2':sl.LOG_LOCAL2,
				'LOCAL3':sl.LOG_LOCAL3, 'LOCAL4':sl.LOG_LOCAL4,
				'LOCAL5':sl.LOG_LOCAL5, 'LOCAL6':sl.LOG_LOCAL6,
				'LOCAL7':sl.LOG_LOCAL7, 'LPR':sl.LOG_LPR,
				'MAIL':sl.LOG_MAIL, 'NEWS':sl.LOG_NEWS,
				'SYSLOG':sl.LOG_SYSLOG, 'USER':sl.LOG_USER,
				'UUCP':sl.LOG_UUCP
			}
			self.set("logging","log facility",str(facilities[facility]))
 		except ConfigParser.NoOptionError: self.set("logging","log facility",str(sl.LOG_DAEMON))
		except KeyError: raise ValueError, "Invalid log facility %s"%facility

		if not self.has_section("load controller"): self.add_section("load controller")
 		try:
			try:
				load = self.getfloat("load controller","target load")
			except ValueError:
				load = self.get("load controller","target load")
				raise ValueError, "Invalid value %s for target load"%load
			if load < 0.1 or load > 10:
				raise ValueError, "Invalid value %s for target load"%load
 		except ConfigParser.NoOptionError: self.set("load controller","target load",str(1.0))

		if not self.has_section("socket interface"): self.add_section("socket interface")
		try: sockaddr = self.get("socket interface","socket")
		except ConfigParser.NoOptionError:
			self.set("socket interface","socket",os.path.join(self.get_data_dir(),"socket"))



	
def get(config_filename=None):
	return MetadataServiceConfig(config_filename)
